#include <assert.h>

#include <string>
#include <iostream>
#include <vector>
#include <deque>
#include <set>

#include "wreath.h"
#include "rtree.h"

using namespace std;

#include <gmpxx.h>

bint count_rtrees(int n, int r)
{
  if (n == 0)
    return r;
  bint lastk = count_rtrees(n-1, r);
  bint sum;
  for (int c = 1; c <= r; c++)
    {
      if (r % c != 0)
	continue;
      for (int d = 1; d <= c; d++)
	{
	  if (c % d != 0)
	    continue;
	  sum += bpow(lastk, r/c) * (mu(c/d)*d*d);
	}
    }
  return sum / r;
}

bool rpequal (const rtree* a, const rtree* b)
{
  if (a == b)
    return true;
  if (!a || !b)
    return false; // one is null, hence other is not
  if (a->r != b->r || a->root != b->root)
    return false;
  for (int i = 0; i < a->r; i++)
    {
      if (!rpequal(a->kids[i], b->kids[i]))
	return false;
    }
  return true;
}

int rtree::get_d () const
{
  if (d)
    return d;
  for (int i = 1; i < r; i++)
    {
      if (r % i != 0)
	continue;
      // attempt to cycle by i
      bool ok = true;
      for (int j = 0; j < r - i; j++)
	{
	  if (!rpequal(kids[j], kids[j+i]))
	    {
	      ok = false;
	      break;
	    }
	}
      if (ok)
	{
	  // the subgroup of Z/rZ generated by i is isomorphic to Z/(r/i)Z
	  d = r/i;
	  return d;
	}
    }
  // Z/1Z ~= <1> trivially stabilizes, and no other one worked.
  d = 1;
  return d;
}

const bint& rtree::get_dim () const
{
  if (dim != 0)
    return dim;
  if (isleaf())
    {
      dim = 1;
      return dim;
    }
  dim = 1;
  for (int i = 0; i < r; i++)
    {
      assert(kids[i]);
      dim *= kids[i]->get_dim(); // tensored
    }
  // induce: index of G wr Z_d in G wr Z_r is r/d
  dim *= (r/d);
  return dim;
}

ostream& rtree::print(ostream& os) const
{
  if (isleaf())
    os << "(" << root << ")";
  else
    {
      os << "(" << root;
      for (int i = 0; i < r; i++)
	{
	  assert(kids[i]);
	  os << " ";
	  kids[i]->print(os);
	}
      os << ")";
    }
  return os;
}

ostream& rtree::print_parsetree (ostream& os) const
{
  if (isleaf())
    os << " ." << root << ". ";
  else
    {
      os << "( ." << root << ". ";
      for (int i = 0; i < r; i++)
	{
	  assert(kids[i]);
	  os << " ";
	  kids[i]->print_parsetree(os);
	}
      os << ") ";
    }
  return os;
}

// want to iterate over all length l vectors of {0,1,...,n-1} which are not
// cyclic permutations of each other.

typedef deque<int> cvec;

template<class T> static void cycle_right(T& v)
{
  v.push_back(v.front());
  v.pop_front();
}

template<class T> static void cycle_left(T& v)
{
  v.push_front(v.back());
  v.pop_back();
}

class cvec_gen {
  int n;
  int len;
  bool alldone;
  cvec v;
  set<cvec> seen;

public:
  cvec_gen (int n_, int len_) 
    : n(n_), len(len_), alldone(false), v(len,0) { };

  ~cvec_gen () { };
  
  const cvec& get (void) const
  {
    return v;
  }

  bool done (void) const
  {
    return alldone;
  }

public:
  void next (void);
};

// return true if overflow
static bool incv (cvec::iterator i, cvec::iterator end, int base)
{
  if (i == end)
    return true;
  assert(*i < base);
  if (++*i >= base)
    {
      *i = 0;
      return incv(++i, end, base);
    }
  else
    return false;
}

void cvec_gen::next (void)
{
  // save what we've seen
  seen.insert(v);

  while (!(alldone = incv(v.begin(), v.end(), n)))
    {
      cvec v1(v);
      bool ok = true;
      for (int i = 0; i < len; i++)
	{
	  if (seen.find(v1) != seen.end())
	    {
	      ok = false;
	      break;
	    }
	  cycle_left(v1);
	}
      if (ok)
	break;
    }
}


size_t estimate_rtree_size(int r) {
  return sizeof(rtree) + r * sizeof(rtree *);
}

// rtree generation

const int max_depth = 10;

vector<rtree> all_rtrees[max_depth];

int all_r; // all_rtrees depends on r, so it only makes sense for r to be fixed

void clear_all_rtrees(void)
{
  for (int i = 0; i < max_depth; i++)
    all_rtrees[i].clear();
}

#if 0
// base r addition.  lsb stored first
// return true if carry out

bool inc_base_r(int *num, int len, int base)
{
  assert(len >= 1);
  assert(base >= 2);
  int carry = 1;
  for (int i = 0; i < len && carry; i++)
    {
      assert(num[i] < base);
      num[i] += carry;
      if (num[i] == base)
	{
	  num[i] = 0;
	  carry = 1;
	}
      else
	carry = 0;
    }
  return carry;
}
#if 0
bool inc_base_r(int *num, int len, int base)
{
  assert(base >= 2);
  if (len == 0)
    return true;
  else
    {
      assert(*num < base);
      if (++*num >= base)
	return inc_base_r(num+1, len-1, base);
      else
	return false;
    }
}
#endif      
#endif

void generate_all_rtrees(int depth, int r)
{
  if (r != all_r)
    {
      clear_all_rtrees();
      all_r = r;
    }
  if (!all_rtrees[depth].empty())
    return; // already done
  if (depth == 0)
    {
      for (int i = 0; i < r; i++)
	{
	  rtree rt(r);
	  rt.makeleaf(i);
	  all_rtrees[depth].push_back(rt);
	}
    }
  else
    {
      generate_all_rtrees(depth - 1, r);

      bint needed = count_rtrees(depth, r);
      bint done = 0;

      cerr << "Generating " << needed << " rtrees with n=" << depth
	   << ", r=" << r << endl;
      cerr << "This will take about " << needed * estimate_rtree_size(r)
	   << " bytes" << endl;

      int n_subtrees = all_rtrees[depth - 1].size();
      for (cvec_gen cvg(n_subtrees, r); !cvg.done(); cvg.next())
	{
	  rtree rt(r);
	  const cvec& v = cvg.get();
	  for (int i = 0; i < r; i++)
	    rt.kids[i] = &all_rtrees[depth - 1][v[i]];
	  int d = rt.get_d();
	  for (int j = 0; j < d; j++)
	    {
	      rt.root = j * (r/d);
	      all_rtrees[depth].push_back(rt);
	      if (++done % 100 == 0)
		cerr << "Done with " << done << " of " << needed << "    \r";
	    }
	}
      cerr << endl;
    }
}

bint wr_order(int n, int r)
{
  if (n == 1)
    return r;
  else
    return bpow(wr_order(n-1, r), r) * r;
}
